// Datei: digital.cpp
// Simulation einer digitalen Schaltung

# include <stdio.h>


//***** Klassen-Deklarationen *****
class Ereignis
  {
    public:
      Ereignis (double z, int n, int w);

      double Zeit;
      int    Netz;
      int    Wert;
  };

class Ereignisliste
  {
    public:
      Ereignisliste ();

      void addiereEreignis (Ereignis e);
      Ereignis gibNaechstesEreignis (); // loescht zugleich Ereignis
      int istLeer ();

    private:
      Ereignis Liste[100];
      int Listenlaenge;
  };

class Gatter
  {
    public:
      Gatter ();
      Gatter (int t, double d, int o, int i1, int i2);

      void reagiereAufEreignis (Ereignis e);

    private:
      int    typ;
      double delay;
      int    output, input1, input2;
  };


//***** globale Variablen und Konstanten *****
//+++++ System-Variablen +++++
double Zeit;                 // aktuelle Simulationszeit
double max_Zeit;             // max. Simulationsdauer
Ereignisliste EventQueue;    // zentrale Ereignisliste

//+++++ Modell-Variablen und -Konstanten +++++
const max_Netz   = 100;
const max_Gatter = 100;

int    N[max_Netz+1];
int    Netzzahl = 0;
Gatter G[max_Gatter+1];
int    Gatterzahl = 0;

//+++++ Gatter-Typ-Konstanten +++++
const int INV  = 1;
const int AND  = 2;
const int OR   = 3;
const int NAND = 4;
const int NOR  = 5;


//***** Klassen-Implementierungen *****
//+++++ Klasse Ereignis +++++
Ereignis::Ereignis (double z = 0.0, int n = 0, int w = 0)
  {
    Zeit = z;
    Netz = n;
    Wert = w;
  }


//+++++ Klasse Ereignisliste +++++
Ereignisliste::Ereignisliste ()
  {
    Listenlaenge = 0;
  }

void Ereignisliste::addiereEreignis (Ereignis e)
  {
    Listenlaenge = Listenlaenge + 1;
    Liste[Listenlaenge] = e;
  }

Ereignis Ereignisliste::gibNaechstesEreignis ()
  {
    Ereignis e;
    double   min;
    int      i, pos_min;

    if (istLeer())
        return (Ereignis(0.0));
      else
        {
          // naechstes Ereignis suchen
          min = Liste[1].Zeit;
          pos_min = 1;
          for (i=2; i<=Listenlaenge; i=i+1)
            if (Liste[i].Zeit < min)
              {
                min = Liste[i].Zeit;
                pos_min = i;
              }
          e = Liste[pos_min];

          // Ereignis aus Liste entfernen
          for (i=pos_min; i<Listenlaenge; i=i+1)
            Liste[i] = Liste[i+1];
          Listenlaenge = Listenlaenge - 1;

          return (e);
        }
  }

int Ereignisliste::istLeer ()
  {
    return (Listenlaenge == 0);
  }


//+++++ Klasse Gatter +++++
Gatter::Gatter ()
  {
    typ    = 0;
    delay  = 0.0;
    output = 0;
    input1 = 0;
    input2 = 0;
  }

Gatter::Gatter (int t, double d, int o, int i1, int i2)
  {
    typ    = t;
    delay  = d;
    output = o;
    input1 = i1;
    input2 = i2;
  }

void Gatter::reagiereAufEreignis (Ereignis e)
  {
    int neu;

    if ((e.Netz > 0) && ((e.Netz == input1) || (e.Netz == input2)))
      {
        switch (typ)
          {
            case   INV: neu = ! N[input1]; break;
            case   AND: neu = N[input1] && N[input2]; break;
            case    OR: neu = N[input1] || N[input2]; break;
            case  NAND: neu = !(N[input1] && N[input2]); break;
            case   NOR: neu = !(N[input1] || N[input2]); break;
          }
        EventQueue.addiereEreignis(Ereignis(Zeit+delay,output,neu));
      }
  }


//***** Hauptprogramm *****
void Initialisierung ()
  { int i;

    // alle Netze auf 0 initialisieren
    for (i=0; i<=max_Netz; i=i+1)
      N[i] = 0;

    // Schaltung definieren
    G[1] = Gatter ( AND, 10.0, 5, 1, 3);
    G[2] = Gatter ( AND, 10.0, 6, 2, 4);
    G[3] = Gatter ( INV,  6.5, 4, 1, 0);
    G[4] = Gatter ( INV,  6.5, 3, 2, 0);
    G[5] = Gatter (  OR, 12.5, 7, 5, 6);
    Gatterzahl = 5;
    Netzzahl   = 7;

    // Initialisierung der Eingaenge
    EventQueue.addiereEreignis (Ereignis(-100, 1, 0));
    EventQueue.addiereEreignis (Ereignis(-100, 2, 0));

    // Input-Stimuli definieren
    EventQueue.addiereEreignis (Ereignis( 100, 2, 1));
    EventQueue.addiereEreignis (Ereignis( 200, 2, 0));
    EventQueue.addiereEreignis (Ereignis( 200, 1, 1));
    EventQueue.addiereEreignis (Ereignis( 300, 2, 1));
    EventQueue.addiereEreignis (Ereignis( 400, 1, 0));
    EventQueue.addiereEreignis (Ereignis( 400, 2, 0));

    // addiere Pseudoereignis fuer t=0
    EventQueue.addiereEreignis (Ereignis(0,0,1));

    max_Zeit = 500;

    // Tabellenkopf fuer Ergebnis-Ausgabe
    printf ("           n n n n n n n\n");
    printf ("     Zeit  1 2 3 4 5 6 7\n");
    printf ("---------  -------------\n");
  }


void main ()
  {
    Ereignis e;
    int i;

    Initialisierung();

    e = EventQueue.gibNaechstesEreignis();
    while ((e.Zeit <= max_Zeit) && !(EventQueue.istLeer()))
      {
        Zeit = e.Zeit;

        if (N[e.Netz] != e.Wert)
          { // Ereignis bewirkt einen Signalwechsel
            N[e.Netz] = e.Wert;
            if (Zeit >= 0)
              { // Status-Anzeige
                printf ("%9.1lf ", Zeit);
                for (i=1; i<=Netzzahl; i=i+1)
                  printf (" %d", N[i]);
                printf ("\n");
              }
          }

        // lasse Gatter darauf reagieren
        // (wg. Init. auch ohne Wechsel)
        for (i=1; i<=Gatterzahl; i=i+1)
          G[i].reagiereAufEreignis(e);

        e = EventQueue.gibNaechstesEreignis();
      }
  }