// Datei: formel.cpp
// Formel-Uebersetzer

# include <stdlib.h>
# include <stdio.h>


//***** Klassen-Deklarationen *****

typedef class Parsebaum *Baumzeiger;

class Parsebaum
  {
    public:
      Parsebaum (int op, Baumzeiger op1, Baumzeiger op2, int w);

      void  anzeigen ();
      float WertFuer (float x);

    private:
      int        Operator;
      Baumzeiger Operand1, Operand2;
      int        Wert;
  };

class Scanner
  {
    public:
      Scanner ();

      void init (char* s);
      int  gibNaechstesSymbol ();
      int  gibZahl ();
      int  SymbolPosition ();

    private:
      char* String;
      int   Position;
      int   PositionSymbol;
      int   Wert;
  };

class Parser
  {
    public:
      Baumzeiger parse (char* s);

    private:
      void leseNaechstesSymbol ();
      void meldeFehler (char* s);

      Baumzeiger parseAusdruck ();
      Baumzeiger parseTerm ();
      Baumzeiger parseVorzeichenfaktor ();
      Baumzeiger parseFaktor ();

      char*   Term;
      Scanner scanner;
      int     aktSymbol;
  };


//***** Token-Konstanten *****
const UNBEKANNT  = 0;
const PLUS       = 1;
const MINUS      = 2;
const MAL        = 3;
const GETEILT    = 4;
const KLAMMERAUF = 5;
const KLAMMERZU  = 6;
const VARIABLE   = 7;
const ZAHL       = 8;
const ENDE       = 9;

const NEGATION   = 10;



//***** Klassen-Implementierungen *****
//+++++ Klasse Parsebaum +++++
Parsebaum::Parsebaum (int op, Baumzeiger op1, Baumzeiger op2, int w)
// Konstruktor
  {
    Operator = op;
    Operand1 = op1;
    Operand2 = op2;
    Wert     = w;
  }


void Parsebaum::anzeigen ()
// gibt den Parsebaum als Term auf dem Bildschirm aus
  {
    switch (Operator)
      {
        case PLUS:     printf ("(");
                       (*Operand1).anzeigen ();
                       printf ("+");
                       (*Operand2).anzeigen ();
                       printf (")");
                       break;
        case MINUS:    printf ("(");
                       (*Operand1).anzeigen ();
                       printf ("-");
                       (*Operand2).anzeigen ();
                       printf (")");
                       break;
        case MAL:      printf ("(");
                       (*Operand1).anzeigen ();
                       printf ("*");
                       (*Operand2).anzeigen ();
                       printf (")");
                       break;
        case GETEILT:  printf ("(");
                       (*Operand1).anzeigen ();
                       printf ("/");
                       (*Operand2).anzeigen ();
                       printf (")");
                       break;
        case NEGATION: printf ("-");
                       (*Operand1).anzeigen ();
                       break;
        case VARIABLE: printf ("x");
                       break;
        case ZAHL:     printf ("%d", Wert);
      }
  }


float Parsebaum::WertFuer (float x)
// gibt den Wert des Ausdrucks fuer Variable = x zurueck
  {
    float nenner;

    switch (Operator)
      {
        case PLUS:     return ((*Operand1).WertFuer(x) +
                               (*Operand2).WertFuer(x));
        case MINUS:    return ((*Operand1).WertFuer(x) -
                               (*Operand2).WertFuer(x));
        case MAL:      return ((*Operand1).WertFuer(x) *
                               (*Operand2).WertFuer(x));
        case GETEILT:  nenner = (*Operand2).WertFuer(x);
                       if (nenner == 0.0)
                         {
                           printf ("Fehler: Division durch 0!");
                           exit (1);
                         }
                       return ((*Operand1).WertFuer(x) / nenner);
        case NEGATION: return (-(*Operand1).WertFuer(x));
        case VARIABLE: return (x);
        case ZAHL:     return (Wert);
      }
  }

  
//+++++ Klasse Scanner +++++
Scanner::Scanner ()
// Default-Konstruktor
  {
    String   = "";
    Position = 0;
  }


void Scanner::init (char* s)
// initialisiert den Scanner mit dem Input-String
  {
    String   = s;
    Position = 0;
  }


int Scanner::gibNaechstesSymbol ()
// scannt den Input-String und gibt das naechste Symbol zurueck
  {
    char aktChar;
    int  symbol;

    // naechstes Zeichen einlesen
    while ((aktChar = String[Position]) == ' ')
      Position = Position + 1;
    PositionSymbol = Position;

    if (String[Position] == 0)
      return (ENDE);

    Position = Position + 1;

    // Symbol auswerten
    if (aktChar == '+')
      {return (PLUS);}

    if (aktChar == '-')
      {return (MINUS);}

    if (aktChar == '*')
      {return (MAL);}

    if (aktChar == '/')
      {return (GETEILT);}

    if (aktChar == '(')
      {return (KLAMMERAUF);}

    if (aktChar == ')')
      {return (KLAMMERZU);}

    if ((aktChar == 'x') || (aktChar == 'X'))
      {return (VARIABLE);}

    if ((aktChar >= '0') && (aktChar <= '9'))
      {
        Wert = aktChar - '0';
        while ((String[Position] >= '0') &&
               (String[Position] <= '9'))
          {
            Wert = Wert*10 + (String[Position] - '0');
            Position = Position + 1;
          }
        return (ZAHL);
      }

    return (UNBEKANNT);
  }


int Scanner::gibZahl ()
// gibt den letzten eingelesenen Zahlenwert zurueck
  {return (Wert);}


int Scanner::SymbolPosition ()
// gibt die Anfangsposition des letzten Symbols zurueck
  {return (PositionSymbol);}


//+++++ Klasse Parser +++++
void Parser::leseNaechstesSymbol ()
// ruft das naechste Symbol vom Scanner ab
  {aktSymbol = scanner.gibNaechstesSymbol ();}


void Parser::meldeFehler (char* s)
// meldet einen Fehler und bricht die Ausfuehrung ab
  {
    int i;

    printf ("%s\n", Term);
    for (i=0; i<scanner.SymbolPosition(); i=i+1)
      printf (" ");
    printf ("^\n\n");
    printf ("Fehler: %s\n", s);
    exit (1);
  }


Baumzeiger Parser::parse (char* s)
// startet den Parse-Lauf fuer die gesamte Formel
  {
    Baumzeiger baum;

    Term = s;
    scanner.init (s);

    leseNaechstesSymbol ();
    baum = parseAusdruck ();
    if (aktSymbol != ENDE)
      meldeFehler ("unerwartetes Zeichen");
    return (baum);
  }


Baumzeiger Parser::parseAusdruck ()
// liest einen Ausdruck ein
// Ausdruck := Term {("+"|"-") Term}
  {
    int op;
    Baumzeiger baum, op2;

    baum = parseTerm ();
    while ((aktSymbol == PLUS) | (aktSymbol == MINUS))
      {
        op = aktSymbol;
        leseNaechstesSymbol ();
        op2 = parseTerm ();
        baum = new Parsebaum (op, baum, op2, 0);
      }
    return (baum);
  }


Baumzeiger Parser::parseTerm ()
// liest einen Term ein
// Term := Vorzeichenfaktor {("*"|"/") Vorzeichenfaktor}
  {
    int op;
    Baumzeiger baum, op2;

    baum = parseVorzeichenfaktor ();
    while ((aktSymbol == MAL) | (aktSymbol == GETEILT))
      {
        op = aktSymbol;
        leseNaechstesSymbol ();
        op2 = parseVorzeichenfaktor ();
        baum = new Parsebaum (op, baum, op2, 0);
      }
    return (baum);
  }


Baumzeiger Parser::parseVorzeichenfaktor ()
// liest einen vorzeichenbehafteten Faktor ein
// Vorzeichenfaktor := ["-"] Faktor
  {
    int vorzeichen = 0;
    Baumzeiger baum;

    if (aktSymbol == MINUS)
      {
        vorzeichen = 1;
        leseNaechstesSymbol ();
      }
    baum = parseFaktor ();
    if (vorzeichen == 1)
      baum = new Parsebaum (NEGATION, baum, NULL, 0);
    return (baum);
  }


Baumzeiger Parser::parseFaktor ()
// liest einen Faktor ein
// Faktor := "x" | Zahl | "(" Ausdruck ")"
  {
    Baumzeiger baum;

    if (aktSymbol == VARIABLE)
      baum = new Parsebaum(VARIABLE, NULL, NULL, 0);
    else if (aktSymbol == ZAHL)
      baum = new Parsebaum(ZAHL, NULL, NULL, scanner.gibZahl());
    else if (aktSymbol == KLAMMERAUF)
      {
        leseNaechstesSymbol ();
        baum = parseAusdruck ();
        if (aktSymbol != KLAMMERZU)
          meldeFehler ("')' erwartet");
      }
    else if (aktSymbol == UNBEKANNT)
      meldeFehler ("unbekanntes Symbol");
    else if (aktSymbol == ENDE)
      meldeFehler ("unerwartetes Ende");
    else
      meldeFehler ("unerwartetes Symbol");

    leseNaechstesSymbol ();
    return (baum);
  }



//***** Hauptprogramm *****
void main ()
  {
    // Variablen-Definitionen
    char       Term[200];
    Parser     parser;
    Baumzeiger PBaum;
    float      x;

    //Term einlesen
    printf ("Term eingeben: ");
    scanf ("%[^\n]", Term);
    printf ("\n");

    // Term parsen und Ergebnis anzeigen
    PBaum = parser.parse (Term);
    (*PBaum).anzeigen();
    printf ("\n\n");

    // Wertetabelle fuer Term berechnen
    printf ("   x    f(x)\n");
    printf ("----  ------\n");
    for (x=0.0; x<=10.0; x=x+1)
      printf ("%4.1f  %6.1f\n", x, (*PBaum).WertFuer(x));

  }
