// Datei: schlang2.cpp
// ereignisorientierte Warteschlangensimulation
// ein Lauf mit mehreren Batches

# include <stdio.h>
# include "e:\compiler\tcwin\prog\zgen.cpp"


//***** Klassen-Deklarationen *****

class Ereignis
  {
    public:
      Ereignis ();
      Ereignis (const double z, const int a);

      double Zeit;
      int    Aktion;
  };

class Ereignisliste
  {
    public:
      Ereignisliste ();

      void addiereEreignis (Ereignis e);
      Ereignis gibNaechstesEreignis (); // loescht zugleich Ereignis
      void leeren ();
      int istLeer ();

    private:
      Ereignis Liste[100];
      int Listenlaenge;
  };

class Kundengenerator
  {
    public:
      void reagiereAufEreignis (Ereignis e);

    private:
      void generiereNaechstenKunden ();
  };

class Bediensystem
  {
    public:
      Bediensystem();

      void RuecksetzenSchlange ();
      void RuecksetzenStatistik (double start);
      void reagiereAufEreignis (Ereignis e);

      double mittlere_Schlangenlaenge ();
      double mittlere_Aufenthaltsdauer ();
      double mittlere_Auslastung ();
      long   Gesamtzahl_Kunden ();
      long   aktueller_Kunde ();
      long   letzter_Kunde ();
      long   aktuelle_Schlangenlaenge ();

    private:
      void AnkunftKunde ();
      void BeginnBedienung ();
      void EndeBedienung ();
      void StatistikMitschreiben ();

      // Variablen Systemstatus
      long letzte_KundenNr;         // Nr. letzten angekom. Kunden
      long bedienter_Kunde;         // Nr. des Ersten der Schlange

      // Variablen Statistik
      double ZeitLetztesEreignis;   // Zeitpunkt letztes Ereignis
      long   Kundenzahl;            // Anzahl der erfassten Kunden
      double Dauer;                 // Zeitdauer fuer Statistik
      double Summe_Bedienung;       // Gesamtdauer Bedienung
      double Summe_Schlangenlaenge; // Summe aller Schlangenlaengen
  };


//***** globale Variablen *****
double Zeit;                 // aktuelle Simulationszeit
Ereignisliste EventQueue;    // zentrale Ereignisliste
Zufallsgenerator z_gen;      // zentraler Zufallsgenerator

//***** Ereignis-Konstanten *****
const Ereignis_Ankunft_Kunde  = 1;
const Ereignis_Ende_Bedienung = 2;
const Pseudo_Ereignis         = 3;


//***** Klassen-Implementierungen *****
//+++++ Klasse Ereignis +++++
Ereignis::Ereignis ()
// Default-Konstruktor
  {
    Zeit   = 0.0;
    Aktion = 0;
  }

Ereignis::Ereignis (const double z, const int a)
// Konstruktor mit Parametern
  {
    Zeit   = z;
    Aktion = a;
  }


//+++++ Klasse Ereignisliste +++++
Ereignisliste::Ereignisliste ()
// Konstruktor
  {Listenlaenge = 0;}

void Ereignisliste::leeren ()
// Ereignisliste leeren
  {Listenlaenge = 0;}

void Ereignisliste::addiereEreignis (Ereignis e)
// Ereignis an das Ende der Liste anhaengen
  {
    Listenlaenge = Listenlaenge + 1;
    Liste[Listenlaenge] = e;
  }

Ereignis Ereignisliste::gibNaechstesEreignis ()
// zeitlich naechstes Ereignis durch sequentielle Suche bestimmen
// gefundenes Ereignis aus Liste loeschen und zurueckgeben
  {
    Ereignis e;
    double min;
    int    i, pos_min;

    // naechstes Ereignis suchen
    min = Liste[1].Zeit;
    pos_min = 1;
    for (i=2; i<=Listenlaenge; i=i+1)
      if (Liste[i].Zeit < min)
        {
          min = Liste[i].Zeit;
          pos_min = i;
        }
    e = Liste[pos_min];

    // Ereignis aus Liste entfernen
    Liste[pos_min] = Liste[Listenlaenge];
    Listenlaenge = Listenlaenge - 1;

    return (e);
  }

int Ereignisliste::istLeer ()
// gibt TRUE zurueck, wenn Ereignisliste leer ist
  {return (Listenlaenge == 0);}


//+++++ Klasse Kundengenerator +++++
void Kundengenerator::reagiereAufEreignis (Ereignis e)
// reagiert auf passendes Ereignis durch Unterprogrammaufruf
  {
    if (e.Aktion == Ereignis_Ankunft_Kunde)
      generiereNaechstenKunden ();
  }

void Kundengenerator::generiereNaechstenKunden ()
// erzeugt das Ereignis Ankunft naechster Kunde
  {
    double z = Zeit + z_gen.exponentialverteilt(1.0/3.0);
    EventQueue.addiereEreignis(Ereignis(z, Ereignis_Ankunft_Kunde));
  }


//+++++ Klasse Bediensystem +++++
void Bediensystem::RuecksetzenSchlange ()
// definiert die Schlange als leer
  {
    letzte_KundenNr = 0;
    bedienter_Kunde = 0;
  }

void Bediensystem::RuecksetzenStatistik (double start = 0.0)
// initialisiert die Variablen fuer die statistische Auswertung
  {
    ZeitLetztesEreignis   = start;
    Kundenzahl            = 0;
    Dauer                 = 0.0;
    Summe_Bedienung       = 0.0;
    Summe_Schlangenlaenge = 0.0;
  }

Bediensystem::Bediensystem ()
// Konstruktor
  {
    RuecksetzenSchlange ();
    RuecksetzenStatistik ();
  }

void Bediensystem::reagiereAufEreignis (Ereignis e)
// reagiert auf passendes Ereignis durch Unterprogrammaufruf
  {
    StatistikMitschreiben ();

    switch (e.Aktion)
      {
        case Ereignis_Ankunft_Kunde : AnkunftKunde();  break;
        case Ereignis_Ende_Bedienung: EndeBedienung(); break;
      }
  }

void Bediensystem::AnkunftKunde ()
// verarbeitet das Ereignis Ankunft Kunde
  {
    Kundenzahl = Kundenzahl + 1;
    letzte_KundenNr = letzte_KundenNr + 1;

    if (bedienter_Kunde == 0)
      {
        bedienter_Kunde = letzte_KundenNr;
        BeginnBedienung();
      }
  }

void Bediensystem::BeginnBedienung ()
// bearbeitet den Beginn einer Bedienung
  {
    double z = Zeit + z_gen.exponentialverteilt(1.0/2.5);
    EventQueue.addiereEreignis(Ereignis(z,Ereignis_Ende_Bedienung));
  }

void Bediensystem::EndeBedienung ()
// verarbeitet das Ereignis Ende Bedienung
  {
    if (bedienter_Kunde == letzte_KundenNr)
        bedienter_Kunde = 0;
      else
        {
          bedienter_Kunde = bedienter_Kunde + 1;
          BeginnBedienung();
        }
  }

void Bediensystem::StatistikMitschreiben ()
// aktualisiert die Variablen fuer die statistische Auswertung
  {
    double delta = Zeit - ZeitLetztesEreignis;

    Dauer = Dauer + delta;
    if (bedienter_Kunde > 0)
      Summe_Bedienung = Summe_Bedienung + delta;
    Summe_Schlangenlaenge = Summe_Schlangenlaenge +
                              aktuelle_Schlangenlaenge() * delta;
    ZeitLetztesEreignis = Zeit;
  }

long Bediensystem::aktuelle_Schlangenlaenge ()
// berechnet die aktuelle Schlangenlaenge und gibt sie zurueck
  {
    if (bedienter_Kunde == 0)
        return (0);
      else
        return (letzte_KundenNr - bedienter_Kunde + 1);
  }

long Bediensystem::aktueller_Kunde ()
// gibt die Nr. des aktuell bedienten Kunden zurueck (sonst 0)
  {return(bedienter_Kunde);}

long Bediensystem::letzter_Kunde ()
// gibt die Nr. des letzten Kunden zurueck
  {return(letzte_KundenNr);}

double Bediensystem::mittlere_Schlangenlaenge ()
// gibt die mittl. Schlangenlaenge fuer die lfd. Auswertung zurueck
  {return (Summe_Schlangenlaenge/Dauer);}

double Bediensystem::mittlere_Aufenthaltsdauer ()
// gibt die mittl. Aufenthaltsdauer fuer die lfd. Auswertung zurueck
  {return (Summe_Schlangenlaenge/Kundenzahl);}

double Bediensystem::mittlere_Auslastung ()
// gibt die mittl. Schlangenlaenge fuer die lfd. Auswertung zurueck
  {return (Summe_Bedienung*100.0/Dauer);}

long Bediensystem::Gesamtzahl_Kunden ()
// gibt die Kundenzahl fuer die lfd. Auswertung zurueck
  {return (Kundenzahl);}


//***** Hauptprogramm *****
// ein langer Lauf, aufgeteilt in Batches
void main ()
  {
    // Variablen-Definition
    Kundengenerator k_gen;
    Bediensystem    b_sys;
    Ereignis        aktuelles_Ereignis;
    long            Anzahl_Batches = 100;
    double          Batch_Laenge = 5000.0;
    double          Summe_Schlangenlaengen   = 0.0;
    double          QSumme_Schlangenlaengen  = 0.0;
    double          Summe_Aufenthaltsdauern  = 0.0;
    double          QSumme_Aufenthaltsdauern = 0.0;
    double          Summe_Auslastungen       = 0.0;
    double          QSumme_Auslastungen      = 0.0;
    long            Batch_Nr;
    double          Batch_Ende;
    double          X_quer, S2, delta;


    Zeit = 0.0;

    // generiere Start-Ereignis
    k_gen.reagiereAufEreignis (Ereignis(0,Ereignis_Ankunft_Kunde));

    for (Batch_Nr=1; Batch_Nr<=Anzahl_Batches; Batch_Nr=Batch_Nr+1)
     {
      Batch_Ende = Zeit + Batch_Laenge;
      b_sys.RuecksetzenStatistik(Zeit);
      // addiere Pseudo-Ereignis fuer Batch-Ende
      EventQueue.addiereEreignis (Ereignis(Batch_Ende,
                                           Pseudo_Ereignis));

      do
       {
        aktuelles_Ereignis = EventQueue.gibNaechstesEreignis ();
        Zeit = aktuelles_Ereignis.Zeit;

        if (Zeit <= Batch_Ende)
          {
            // Objekten Ereignis mitteilen
            k_gen.reagiereAufEreignis (aktuelles_Ereignis);
            b_sys.reagiereAufEreignis (aktuelles_Ereignis);
          }
       }
      while ((Zeit < Batch_Ende) && !(EventQueue.istLeer()));

      Summe_Schlangenlaengen   = Summe_Schlangenlaengen +
                                 b_sys.mittlere_Schlangenlaenge();
      QSumme_Schlangenlaengen  = QSumme_Schlangenlaengen +
                                 b_sys.mittlere_Schlangenlaenge() *
                                 b_sys.mittlere_Schlangenlaenge();
      Summe_Aufenthaltsdauern  = Summe_Aufenthaltsdauern +
                                 b_sys.mittlere_Aufenthaltsdauer();
      QSumme_Aufenthaltsdauern = QSumme_Aufenthaltsdauern +
                                 b_sys.mittlere_Aufenthaltsdauer() *
                                 b_sys.mittlere_Aufenthaltsdauer();
      Summe_Auslastungen       = Summe_Auslastungen +
                                 b_sys.mittlere_Auslastung();
      QSumme_Auslastungen      = QSumme_Auslastungen +
                                 b_sys.mittlere_Auslastung() *
                                 b_sys.mittlere_Auslastung();
     }

    // Statistik ueber alle Laeufe anzeigen
    X_quer = Summe_Schlangenlaengen / Anzahl_Batches;
    S2     = QSumme_Schlangenlaengen/Anzahl_Batches-X_quer*X_quer;
    delta  = 1.96 * sqrt(S2 / (Anzahl_Batches - 1));
    printf ("mittlere Schlangenlaenge:  %6.2f (+/- %.2f)\n",
            X_quer, delta);

    X_quer = Summe_Aufenthaltsdauern / Anzahl_Batches;
    S2     = QSumme_Aufenthaltsdauern/Anzahl_Batches-X_quer*X_quer;
    delta  = 1.96 * sqrt(S2 / (Anzahl_Batches - 1));
    printf ("mittlere Aufenthaltsdauer: %6.2f (+/- %.2f)\n",
            X_quer, delta);

    X_quer = Summe_Auslastungen / Anzahl_Batches;
    S2     = QSumme_Auslastungen/Anzahl_Batches - X_quer*X_quer;
    delta  = 1.96 * sqrt(S2 / (Anzahl_Batches - 1));
    printf ("mittlere Auslastung:       %6.2f (+/- %.2f)\n",
            X_quer, delta);
  }
