// Datei: zgen.cpp
// Zufallsgenerator

# include <stdlib.h>
# include <math.h>

class Zufallsgenerator
  {
    public:
      Zufallsgenerator (long x, long y);
      void setzeSeed (long s); // Startwert fuer Zufallsgenerator

      double Zufallszahl();                         // aus (0..1)
      long   BasisZufallszahl();                    // aus (0..m)
      int    gleichverteilt (int x1, int x2);       // aus [x1..x2]
      double gleichverteilt (double x1, double x2); // aus (x1..x2)
      double exponentialverteilt (double lambda);
      double standardnormalverteilt ();
      double normalverteilt (double my, double s);
      int    tabellenverteilt (double feld[10]);

    private:
      long seed; // aktueller interner Zufallswert
      long a, m; // Parameter des Zufallsgenerators
      long q, r; // Hilfsgroessen zur Berechnung
  };

Zufallsgenerator::Zufallsgenerator (long x=40629, long y=2147483399)
  {
    randomize();
    seed = random(30000) + 1;
    a = x;
    m = y;
    q = m / a;
    r = m % a;
  }

void Zufallsgenerator::setzeSeed (long s)
  {seed = s;}

long Zufallsgenerator::BasisZufallszahl ()
  {
    seed = a * (seed % q) - (seed / q) * r;
    // seed = a * (seed % (m/a)) - (seed / (m/a)) * (m%a);
    if (seed < 0)
      seed = seed + m;

    return (seed);
  }

double Zufallsgenerator::Zufallszahl ()
  {
    seed = a * (seed % q) - (seed / q) * r;
    // seed = a * (seed % (m/a)) - (seed / (m/a)) * (m%a);
    if (seed < 0)
      seed = seed + m;

    return (double) seed / m;
  }

int Zufallsgenerator::gleichverteilt (int x1, int x2)
  {return (x1 + (int)((Zufallszahl() * (x2-x1+1))));}

double Zufallsgenerator::gleichverteilt (double x1, double x2)
  {return (x1 + ((Zufallszahl() * (x2-x1))));}

double Zufallsgenerator::exponentialverteilt (double lambda)
  {return (-log(Zufallszahl()) / lambda);}

double Zufallsgenerator::standardnormalverteilt ()
  {
    double summe = -6.0;
    int i;

    for (i=1; i<=12; i=i+1)
      summe = summe + Zufallszahl();
    return (summe);
  }

double Zufallsgenerator::normalverteilt (double my, double s)
  {return (my + (s * (standardnormalverteilt())));}

int Zufallsgenerator::tabellenverteilt (double feld[10])
  {
    double p = Zufallszahl();
    int i = 0;
    while ((p>feld[i]) && (i<9))
      i = i+1;
    if (p <= feld[i])
	return (i);
      else
        return (-1);
  }


